@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 1999-2000 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1999-2005 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileCopy.cc,v 1.89 2002/08/28 12:32:17 lav Exp $ */
+/* $Id: FileCopy.cc,v 1.126.2.1 2007/04/05 13:23:36 lav Exp $ */
 
 /* FileCopyPeer behaviour:
     1) when suspended, does nothing
@@ -46,18 +46,21 @@
 #include "misc.h"
 #include "LsCache.h"
 #include "plural.h"
-#include "OutputJob.h"
+#include "ArgV.h"
 
 #define skip_threshold 0x1000
 #define debug(a) Log::global->Format a
 
 ResDecl rate_period  ("xfer:rate-period","15", ResMgr::UNumberValidate,ResMgr::NoClosure);
 ResDecl eta_period   ("xfer:eta-period", "120",ResMgr::UNumberValidate,ResMgr::NoClosure);
-ResDecl max_redir    ("xfer:max-redirections", "0",ResMgr::UNumberValidate,ResMgr::NoClosure);
+ResDecl max_redir    ("xfer:max-redirections", "5",ResMgr::UNumberValidate,ResMgr::NoClosure);
 
 // FileCopy
 #define super SMTask
 
+#define set_state(s) do { state=(s); \
+   Log::global->Format(11,"FileCopy(%p) enters state %s\n", this, #s); } while(0)
+
 int FileCopy::Do()
 {
    int m=STALL;
@@ -70,17 +73,18 @@ int FileCopy::Do()
    switch(state)
    {
    pre_INITIAL:
-      state=INITIAL;
+      set_state(INITIAL);
       m=MOVED;
    case(INITIAL):
       if(remove_target_first && !put->FileRemoved())
 	 return m;
       remove_target_first=false;
-      if(put->NeedSizeDateBeforehand())
+      if(put->NeedSizeDateBeforehand() || (cont && put->CanSeek()))
       {
 	 if(get->GetSize()==NO_SIZE_YET || get->GetDate()==NO_DATE_YET)
 	 {
 	    put->Suspend();
+	    get->DontStartTransferYet();
 	    get->Resume();
 	    get->WantSize();
 	    if(put->NeedDate())
@@ -88,10 +92,10 @@ int FileCopy::Do()
 	    goto pre_GET_INFO_WAIT;
 	 }
       }
+      if(get->GetSize()==NO_SIZE_YET)
+	 get->WantSize();
       if(get->GetSize()!=NO_SIZE && get->GetSize()!=NO_SIZE_YET)
 	 put->SetEntitySize(get->GetSize());
-      else if(get->GetSize()==NO_SIZE_YET)
-	 get->WantSize();
       if(get->GetDate()!=NO_DATE && get->GetDate()!=NO_DATE_YET)
 	 put->SetDate(get->GetDate());
       else if(get->GetDate()==NO_DATE_YET)
@@ -113,21 +117,30 @@ int FileCopy::Do()
 
       get->Suspend();
       put->Resume();
-      state=PUT_WAIT;
+      set_state(PUT_WAIT);
       m=MOVED;
       /* fallthrough */
    case(PUT_WAIT):
       if(put->Error())
 	 goto put_error;
+      if(put->GetSeekPos()!=FILE_END && get->GetSize()>=0
+      && put->GetSeekPos()>=get->GetSize())
+      {
+	 debug((9,_("copy: destination file is already complete\n")));
+	 if(get->GetDate()!=NO_DATE)
+	    goto pre_CONFIRM_WAIT;  // have to set the date.
+	 goto pre_GET_DONE_WAIT;
+      }
       if(!put->IOReady())
 	 return m;
       /* now we know if put's seek failed. Seek get accordingly. */
       if(get->CanSeek())
 	 get->Seek(put->GetRealPos());
-      get->Resume();
    pre_DO_COPY:
+      get->Resume();
+      get->StartTransfer();
       RateReset();
-      state=DO_COPY;
+      set_state(DO_COPY);
       m=MOVED;
       /* fallthrough */
    case(DO_COPY): {
@@ -137,8 +150,13 @@ int FileCopy::Do()
 	 SetError(put->ErrorText());
 	 return MOVED;
       }
-      if(get->Error())
+      if(get->Error() && get->Size()==0)
       {
+	 if(put->GetPos()>0)
+	 {
+	    put->PutEOF();
+	    Roll(put);
+	 }
       get_error:
 	 SetError(get->ErrorText());
 	 return MOVED;
@@ -170,6 +188,11 @@ int FileCopy::Do()
 	 SetError(_("seek failed"));
 	 return MOVED;
       }
+      if(get->GetSize()>0 && get->GetRealPos()>get->GetSize())
+      {
+	 get->SetSize(NO_SIZE_YET);
+	 get->SetDate(NO_DATE_YET);
+      }
       long lbsize=0;
       if(line_buffer)
 	 lbsize=line_buffer->Size();
@@ -233,20 +256,7 @@ int FileCopy::Do()
       if(b==0) // eof
       {
 	 debug((10,"copy: get hit eof\n"));
-      eof:
-	 put->Resume();
-	 if(line_buffer)
-	 {
-	    line_buffer->Get(&b,&s);
-	    put->Put(b,s);
-	    line_buffer->Skip(s);
-	 }
-	 put->SetDate(get->GetDate());
-	 put->PutEOF();
-	 get->Suspend();
-	 put_eof_pos=put->GetRealPos();
-	 state=CONFIRM_WAIT;
-	 return MOVED;
+	 goto eof;
       }
 
       rate_add=put_buf;
@@ -282,14 +292,15 @@ int FileCopy::Do()
 
 	 // now find eol in line_buffer.
 	 line_buffer->Get(&lb,&ls);
-	 while(ls>0)
+	 const char *eol=0;
+	 if(get->Eof() || get->Error())
+	    eol=lb+ls-1;
+	 else
+	    eol=memrchr(lb,'\n',ls);
+	 if(eol)
 	 {
-	    const char *eol=(const char *)memchr(lb,'\n',ls);
-	    if(!eol)
-	       break;
 	    put->Put(lb,eol-lb+1);
 	    line_buffer->Skip(eol-lb+1);
-	    line_buffer->Get(&lb,&ls);
 	 }
       }
       else
@@ -310,13 +321,33 @@ int FileCopy::Do()
       }
       return m;
    }
+
+   eof:
+      if(line_buffer)
+      {
+	 line_buffer->Get(&b,&s);
+	 put->Put(b,s);
+	 line_buffer->Skip(s);
+      }
+   pre_CONFIRM_WAIT:
+      put->SetSuggestedFileName(get->GetSuggestedFileName());
+      put->SetDate(get->GetDate());
+      if(get->GetSize()!=NO_SIZE && get->GetSize()!=NO_SIZE_YET)
+	 put->SetEntitySize(get->GetSize());
+      put->PutEOF();
+      get->Suspend();
+      put->Resume();
+      put_eof_pos=put->GetRealPos();
+      debug((10,"copy: waiting for put confirmation\n"));
+      set_state(CONFIRM_WAIT);
+      m=MOVED;
    case(CONFIRM_WAIT):
       if(put->Error())
 	 goto put_error;
       /* check if put position is correct */
-      if(put_eof_pos!=put->GetRealPos())
+      if(put_eof_pos!=put->GetRealPos() || put->GetSeekPos()==FILE_END)
       {
-	 state=DO_COPY;
+	 set_state(DO_COPY);
 	 return MOVED;
       }
 
@@ -333,10 +364,10 @@ int FileCopy::Do()
       get->Empty();
       get->PutEOF();
       get->Resume();
-      state=GET_DONE_WAIT;
+      set_state(GET_DONE_WAIT);
       m=MOVED;
       end_time=now;
-      Delete(put); put=0;
+      put->Suspend();
       /* fallthrough */
    case(GET_DONE_WAIT):
       if(get->Error())
@@ -349,12 +380,12 @@ int FileCopy::Do()
       if(!get->Done())
 	 return m;
       debug((10,"copy: get is finished - all done\n"));
-      state=ALL_DONE;
-      Delete(get); get=0;
+      set_state(ALL_DONE);
+      get->Suspend();
       return MOVED;
 
    pre_GET_INFO_WAIT:
-      state=GET_INFO_WAIT;
+      set_state(GET_INFO_WAIT);
       m=MOVED;
    case(GET_INFO_WAIT):
       if(get->Error())
@@ -373,12 +404,12 @@ void FileCopy::Init()
 {
    get=0;
    put=0;
-   state=INITIAL;
+   set_state(INITIAL);
    max_buf=0x10000;
    cont=false;
    error_text=0;
-   rate        =new Speedometer("xfer:rate-period");
-   rate_for_eta=new Speedometer("xfer:eta-period");
+   rate        =MakeRef(new Speedometer("xfer:rate-period"));
+   rate_for_eta=MakeRef(new Speedometer("xfer:eta-period"));
    put_buf=0;
    put_eof_pos=0;
    bytes_count=0;
@@ -393,18 +424,18 @@ void FileCopy::Init()
 FileCopy::FileCopy(FileCopyPeer *s,FileCopyPeer *d,bool c)
 {
    Init();
-   get=s;
-   put=d;
+   get=MakeRef(s);
+   put=MakeRef(d);
    cont=c;
 }
 FileCopy::~FileCopy()
 {
-   Delete(get);
-   Delete(put);
-   Delete(line_buffer);
+   DeleteRef(get);
+   DeleteRef(put);
+   delete line_buffer;
    xfree(error_text);
-   Delete(rate);
-   Delete(rate_for_eta);
+   DeleteRef(rate);
+   DeleteRef(rate_for_eta);
 }
 FileCopy *FileCopy::New(FileCopyPeer *s,FileCopyPeer *d,bool c)
 {
@@ -415,25 +446,15 @@ FileCopy *FileCopy::New(FileCopyPeer *s,FileCopyPeer *d,bool c)
       return res;
    return new FileCopy(s,d,c);
 }
-void FileCopy::Suspend()
+void FileCopy::SuspendInternal()
 {
-   if(get) get->Suspend();
-   if(put) put->Suspend();
-   super::Suspend();
+   if(get) get->SuspendSlave();
+   if(put) put->SuspendSlave();
 }
-void FileCopy::Resume()
+void FileCopy::ResumeInternal()
 {
-   super::Resume();
-   if(state!=PUT_WAIT)
-   {
-      if(get && !(put && put->Size()>=max_buf))
-	 get->Resume();
-   }
-   if(state!=GET_INFO_WAIT)
-   {
-      if(put)
-	 put->Resume();
-   }
+   if(get) get->ResumeSlave();
+   if(put) put->ResumeSlave();
 }
 void FileCopy::Fg()
 {
@@ -454,8 +475,8 @@ void FileCopy::SetError(const char *str)
 {
    xfree(error_text);
    error_text=xstrdup(str);
-   Delete(get); get=0;
-   Delete(put); put=0;
+   DeleteRef(get);
+   DeleteRef(put);
 }
 
 void FileCopy::LineBuffered(int s)
@@ -498,6 +519,8 @@ int FileCopy::GetPercentDone()
       psize=put->range_limit-put->range_start;
    if(psize<0)
       return 100;
+   if(ppos>psize)
+      return -1;
    return percent(ppos,psize);
 }
 const char *FileCopy::GetPercentDoneStr()
@@ -605,7 +628,6 @@ void FileCopy::Kill(int sig)
 #define super Buffer
 void FileCopyPeer::SetSize(off_t s)
 {
-   want_size=false;
    size=s;
    if(seek_pos==FILE_END)
    {
@@ -615,11 +637,10 @@ void FileCopyPeer::SetSize(off_t s)
 	 seek_pos=0;
    }
 }
-void FileCopyPeer::SetDate(time_t d)
+void FileCopyPeer::SetDate(time_t d,int p)
 {
-   want_date=false;
-   date=d;
-   if(date==NO_DATE || date==NO_DATE_YET)
+   date.set(d,p);
+   if(d==NO_DATE || d==NO_DATE_YET)
       date_set=true;
    else
       date_set=false;
@@ -642,14 +663,23 @@ bool FileCopyPeer::Done()
    return false;
 }
 
-FileCopyPeer::FileCopyPeer(direction m)
+void FileCopyPeer::Seek(off_t offs)
+{
+   seek_pos=offs;
+   if(mode==PUT)
+      pos-=in_buffer;
+   Empty();
+   eof=false;
+   broken=false;
+}
+
+FileCopyPeer::FileCopyPeer(dir_t m) : IOBuffer(m)
 {
-   mode=m;
    want_size=false;
    want_date=false;
+   start_transfer=true;
    size=NO_SIZE_YET;
    e_size=NO_SIZE;
-   date=NO_DATE_YET;
    seek_pos=0;
    can_seek=false;
    can_seek0=false;
@@ -663,10 +693,13 @@ FileCopyPeer::FileCopyPeer(direction m)
    use_cache=true;
    write_allowed=true;
    done=false;
+   suggested_filename=0;
+   auto_rename=false;
    Suspend();  // don't do anything too early
 }
 FileCopyPeer::~FileCopyPeer()
 {
+   xfree(suggested_filename);
 }
 
 // FileCopyPeerFA implementation
@@ -694,32 +727,46 @@ int FileCopyPeerFA::Do()
 
    if(Done() || Error())
       return m;
+
+   if(verify)
+   {
+      if(verify->Error())
+	 SetError(verify->ErrorText());
+      if(verify->Done())
+      {
+	 done=true;
+	 m=MOVED;
+      }
+      return m;
+   }
+
+   if(want_size && size==NO_SIZE_YET && (mode==PUT || !start_transfer))
+   {
+      if(session->IsClosed())
+      {
+	 info.file=file;
+	 info.get_size=true;
+	 info.get_time=want_date;
+	 session->GetInfoArray(&info,1);
+	 m=MOVED;
+      }
+      res=session->Done();
+      if(res==FA::IN_PROGRESS)
+	 return m;
+      if(res<0)
+      {
+	 session->Close();
+	 SetSize(NO_SIZE);
+	 return MOVED;
+      }
+      SetSize(info.size);
+      SetDate(info.time);
+      session->Close();
+      return MOVED;
+   }
    switch(mode)
    {
    case PUT:
-      if(want_size)
-      {
-	 if(session->IsClosed())
-	 {
-	    info.file=file;
-	    info.get_size=true;
-	    info.get_time=false;
-	    session->GetInfoArray(&info,1);
-	    m=MOVED;
-	 }
-	 res=session->Done();
-	 if(res==FA::IN_PROGRESS)
-	    return m;
-	 if(res<0)
-	 {
-	    session->Close();
-	    SetSize(NO_SIZE);
-	    return MOVED;
-	 }
-	 SetSize(info.size);
-	 session->Close();
-	 m=MOVED;
-      }
       if(fxp)
       {
 	 if(eof)
@@ -741,6 +788,8 @@ int FileCopyPeerFA::Do()
 	 {
 	    if(date!=NO_DATE && date!=NO_DATE_YET)
 	       session->SetDate(date);
+	    if(e_size!=NO_SIZE && e_size!=NO_SIZE_YET)
+	       session->SetSize(e_size);
 	    res=session->StoreStatus();
 	    if(res==FA::OK)
 	    {
@@ -748,8 +797,9 @@ int FileCopyPeerFA::Do()
 	    fxp_eof:
 	       // FIXME: set date for real.
 	       date_set=true;
-	       done=true;
-	       m=MOVED;
+	       if(!verify)
+		  verify=MakeRef(new FileVerificator(session,file));
+	       return MOVED;
 	    }
 	    else if(res==FA::IN_PROGRESS)
 	       return m;
@@ -810,21 +860,21 @@ bool FileCopyPeerFA::IOReady()
    return session->IOReady();
 }
 
-void FileCopyPeerFA::Suspend()
+void FileCopyPeerFA::SuspendInternal()
 {
    if(fxp && mode==PUT)
       return;
-   session->Suspend();
-   super::Suspend();
+   session->SuspendSlave();
 }
-void FileCopyPeerFA::Resume()
+void FileCopyPeerFA::ResumeInternal()
 {
-   super::Resume();
-   session->Resume();
+   session->ResumeSlave();
 }
 
 const char *FileCopyPeerFA::GetStatus()
 {
+   if(verify)
+      return verify->Status();
    if(!session->IsOpen())
       return 0;
    return session->CurrentStatus();
@@ -858,8 +908,14 @@ void FileCopyPeerFA::OpenSession()
       }
       const char *b;
       int s;
-      if(use_cache && LsCache::Find(session,file,FAmode,&b,&s))
+      int err;
+      if(use_cache && FileAccess::cache->Find(session,file,FAmode,&err,&b,&s))
       {
+	 if(err)
+	 {
+	    SetError(b);
+	    return;
+	 }
 	 size=s;
 	 if(seek_pos>=s)
 	    goto past_eof;
@@ -874,8 +930,20 @@ void FileCopyPeerFA::OpenSession()
 	 return;
       }
    }
+   else // mode==PUT
+   {
+      if(e_size>=0 && size>=0 && seek_pos>=e_size)
+      {
+	 debug((10,"copy dst: seek past eof (seek_pos=%lld, size=%lld)\n",
+		  (long long)seek_pos,(long long)e_size));
+	 eof=true;
+	 if(date==NO_DATE || date==NO_DATE_YET)
+	    return;
+      }
+   }
    session->Open(file,FAmode,seek_pos);
    session->SetFileURL(orig_url);
+   session->SetLimit(range_limit);
    if(mode==PUT)
    {
       if(try_time!=0)
@@ -890,9 +958,9 @@ void FileCopyPeerFA::OpenSession()
    session->RereadManual();
    if(ascii)
       session->AsciiTransfer();
-   if(want_size)
+   if(want_size && size==NO_SIZE_YET)
       session->WantSize(&size);
-   if(want_date && date==NO_DATE_YET)
+   if(want_date && (date==NO_DATE_YET || date.ts_prec>1))
       session->WantDate(&date);
    if(mode==GET)
    {
@@ -905,6 +973,16 @@ void FileCopyPeerFA::OpenSession()
    }
 }
 
+void FileCopyPeerFA::WantSize()
+{
+   struct stat st;
+   if(!strcmp(session->GetProto(),"file")
+   && stat(dir_file(session->GetCwd(),file),&st)!=-1)
+      SetSize(st.st_size);
+   else
+      super::WantSize();
+}
+
 void FileCopyPeerFA::RemoveFile()
 {
    session->Open(file,FA::REMOVE);
@@ -964,9 +1042,6 @@ int FileCopyPeerFA::Get_LL(int len)
 
 	       xfree(file);
 	       file=xstrdup(u.path?u.path:"");
-	       size=NO_SIZE_YET;
-	       date=NO_DATE_YET;
-
 	       xfree(orig_url);
 	       orig_url=xstrdup(loc);
 	    }
@@ -1007,6 +1082,10 @@ int FileCopyPeerFA::Get_LL(int len)
 	       xfree(file);
 	       file=new_file;
 	    }
+
+	    size=NO_SIZE_YET;
+	    date=NO_DATE_YET;
+
 	    try_time=0;
 	    retries=0;
 	    current->Timeout(0); // retry with new location.
@@ -1019,7 +1098,8 @@ int FileCopyPeerFA::Get_LL(int len)
    if(res==0)
    {
       eof=true;
-      LsCache::Add(session,file,FAmode,this);
+      FileAccess::cache->Add(session,file,FAmode,FA::OK,this);
+      SetSuggestedFileName(session->GetSuggestedFileName());
    }
    return res;
 }
@@ -1029,8 +1109,8 @@ int FileCopyPeerFA::Put_LL(const char *buf,int len)
    if(session->IsClosed())
       OpenSession();
 
-   off_t io_at=pos;
-   if(GetRealPos()!=io_at) // GetRealPos can alter pos.
+   off_t io_at=pos; // GetRealPos can alter pos, save it.
+   if(GetRealPos()!=io_at)
       return 0;
 
    if(len==0 && eof)
@@ -1104,6 +1184,7 @@ void FileCopyPeerFA::Init()
    redirections=0;
    can_seek=true;
    can_seek0=true;
+   verify=0;
 }
 
 FileCopyPeerFA::FileCopyPeerFA(FileAccess *s,const char *f,int m)
@@ -1115,7 +1196,7 @@ FileCopyPeerFA::FileCopyPeerFA(FileAccess *s,const char *f,int m)
    session=s;
    reuse_later=true;
    if(FAmode==FA::LIST || FAmode==FA::LONG_LIST)
-      Save(LsCache::SizeLimit());
+      Save(FileAccess::cache->SizeLimit());
 }
 FileCopyPeerFA::~FileCopyPeerFA()
 {
@@ -1127,6 +1208,7 @@ FileCopyPeerFA::~FileCopyPeerFA()
    }
    xfree(file);
    xfree(orig_url);
+   DeleteRef(verify);
 }
 
 FileCopyPeerFA::FileCopyPeerFA(ParsedURL *u,int m)
@@ -1163,7 +1245,7 @@ FileCopyPeerFA *FileCopyPeerFA::New(FileAccess *s,const char *url,int m,bool reu
 // FileCopyPeerFDStream
 #undef super
 #define super FileCopyPeer
-FileCopyPeerFDStream::FileCopyPeerFDStream(FDStream *o,direction m)
+FileCopyPeerFDStream::FileCopyPeerFDStream(FDStream *o,dir_t m)
    : FileCopyPeer(m)
 {
    if(o==0 && m==PUT)
@@ -1186,11 +1268,17 @@ FileCopyPeerFDStream::FileCopyPeerFDStream(FDStream *o,direction m)
    }
    if(stream->usesfd(1))
       write_allowed=false;
+   put_ll_timer=0;
+   if(m==PUT)
+      put_ll_timer=new Timer(0,200);
+   verify=0;
 }
 FileCopyPeerFDStream::~FileCopyPeerFDStream()
 {
-   if(stream && delete_stream)
+   if(delete_stream)
       delete stream;
+   delete put_ll_timer;
+   DeleteRef(verify);
 }
 
 void FileCopyPeerFDStream::Seek_LL()
@@ -1267,6 +1355,27 @@ int FileCopyPeerFDStream::Do()
    int m=STALL;
    if(Done() || Error())
       return m;
+
+   if(verify)
+   {
+      if(verify->Error())
+	 SetError(verify->ErrorText());
+      if(verify->Done())
+      {
+	 if(suggested_filename && stream && stream->full_name && auto_rename)
+	 {
+	    char *dir=alloca_strdup(stream->full_name);
+	    dirname_modify(dir);
+	    debug((5,"copy: renaming `%s' to `%s'\n",stream->full_name,suggested_filename));
+	    if(rename(stream->full_name,dir_file(dir,suggested_filename))==-1)
+	       debug((3,"rename(%s, %s): %s\n",stream->full_name,suggested_filename,strerror(errno)));
+	 }
+	 done=true;
+	 m=MOVED;
+      }
+      return m;
+   }
+
    switch(mode)
    {
    case PUT:
@@ -1274,19 +1383,19 @@ int FileCopyPeerFDStream::Do()
       {
 	 if(eof)
 	 {
+	    getfd(); // give it a chance to create empty file. (FIXME - handle tmp errors)
 	    if(!date_set && date!=NO_DATE && do_set_date)
 	    {
 	       if(date==NO_DATE_YET)
 		  return m;
-	       if(getfd()==-1)
-		  return m;
 	       stream->setmtime(date);
 	       date_set=true;
 	       m=MOVED;
 	    }
 	    if(stream && delete_stream && !stream->Done())
 	       return m;
-	    done=true;
+	    if(!verify)
+	       verify=MakeRef(new FileVerificator(stream));
 	    return MOVED;
 	 }
 	 if(seek_pos==0)
@@ -1296,6 +1405,9 @@ int FileCopyPeerFDStream::Do()
 	 return m;
       while(in_buffer>0)
       {
+	 if(!eof && in_buffer<PUT_LL_MIN
+	 && put_ll_timer && !put_ll_timer->Stopped())
+	    break;
 	 int res=Put_LL(buffer+buffer_ptr,in_buffer);
 	 if(res>0)
 	 {
@@ -1391,7 +1503,8 @@ int FileCopyPeerFDStream::Get_LL(int len)
    if(fd==-1)
       return 0;
 
-   if(want_date || want_size)
+   if((want_date && date==NO_DATE_YET)
+   || (want_size && size==NO_SIZE_YET))
    {
       struct stat st;
       if(fstat(fd,&st)==-1)
@@ -1515,14 +1628,43 @@ int FileCopyPeerFDStream::Put_LL(const char *buf,int len)
 	 return -1;
       }
       if(stream->NonFatalError(errno))
+      {
+	 // in case of full disk, check file correctness.
+	 if(errno==ENOSPC && can_seek)
+	 {
+	    struct stat st;
+	    if(fstat(fd,&st)!=-1)
+	    {
+	       if(st.st_size<seek_base+pos-in_buffer)
+	       {
+		  // workaround solaris nfs bug. It can lose data if disk is full.
+		  if(buffer_ptr>=seek_base+pos-in_buffer-buffer_ptr-st.st_size)
+		     UnSkip(seek_base+pos-in_buffer-st.st_size);
+		  else
+		  {
+		     Empty();
+		     pos=st.st_size;
+		  }
+	       }
+	    }
+	 }
 	 return 0;
+      }
       stream->MakeErrorText();
       SetError(stream->error_text);
       return -1;
    }
    stream->clear_status();
-   if(res==len)
+   if(res==len && skip_cr)
+   {
       res+=skip_cr;
+      // performance gets worse because of writing a single char,
+      // but leaving uncomplete line on screen allows mixing it with debug text.
+      if(write(fd,"\n",1)==1)
+	 res+=1;
+   }
+   if(put_ll_timer)
+      put_ll_timer->Reset();
    return res;
 }
 FgData *FileCopyPeerFDStream::GetFgData(bool fg)
@@ -1534,6 +1676,22 @@ FgData *FileCopyPeerFDStream::GetFgData(bool fg)
    return 0;
 }
 
+void FileCopyPeerFDStream::WantSize()
+{
+   struct stat st;
+   st.st_size=NO_SIZE;
+
+   if(stream->fd!=-1)
+      fstat(stream->fd,&st);
+   else if(stream->full_name)
+      stat(stream->full_name,&st);
+
+   if(st.st_size!=NO_SIZE)
+      SetSize(st.st_size);
+   else
+      super::WantSize();
+}
+
 void FileCopyPeerFDStream::RemoveFile()
 {
    stream->remove();
@@ -1545,6 +1703,8 @@ void FileCopyPeerFDStream::RemoveFile()
 
 const char *FileCopyPeerFDStream::GetStatus()
 {
+   if(verify)
+      return verify->Status();
    return stream->status;
 }
 void FileCopyPeerFDStream::Kill(int sig)
@@ -1570,7 +1730,7 @@ FileCopyPeerDirList::FileCopyPeerDirList(FA *s,ArgV *v)
    : FileCopyPeer(GET)
 {
    session=s;
-   dl=session->MakeDirList(v);
+   dl=MakeRef(session->MakeDirList(v));
    if(dl==0)
       eof=true;
    can_seek=false;
@@ -1579,7 +1739,7 @@ FileCopyPeerDirList::FileCopyPeerDirList(FA *s,ArgV *v)
 
 FileCopyPeerDirList::~FileCopyPeerDirList()
 {
-   Delete(dl);
+   DeleteRef(dl);
    SessionPool::Reuse(session);
 }
 
@@ -1610,77 +1770,113 @@ int FileCopyPeerDirList::Do()
    return MOVED;
 }
 
-FileCopyPeerOutputJob::FileCopyPeerOutputJob(OutputJob *new_o)
-   : FileCopyPeer(PUT)
+// FileVerificator
+void FileVerificator::Init0()
 {
-   o=new_o;
-   DontCopyDate();
+   done=false;
+   error_text=0;
+   verify_buffer=0;
+   verify_process=0;
+   if(!ResMgr::QueryBool("xfer:verify",0))
+      done=true;
 }
-
-int FileCopyPeerOutputJob::Put_LL(const char *buf,int len)
+void FileVerificator::InitVerify(const char *f)
 {
-   off_t io_at=pos;
-   if(GetRealPos()!=io_at) // GetRealPos can alter pos.
-      return 0;
-
-   if(len==0 && eof)
-      return 0;
-
-   if(o->Full())
-      return 0;
-
-   o->Put(buf,len);
-
-   seek_pos+=len; // mainly to indicate that there was some output.
-   return len;
+   if(done)
+      return;
+   ArgV *args=new ArgV(ResMgr::Query("xfer:verify-command",0));
+   args->Append(f);
+   verify_process=new InputFilter(args);
+   verify_process->StderrToStdout();
+   verify_buffer=new IOBufferFDStream(verify_process,IOBuffer::GET);
 }
-
-int FileCopyPeerOutputJob::Do()
+FileVerificator::FileVerificator(const char *f)
 {
-   if(o->Error())
-   {
-      broken=true;
-      return MOVED;
-   }
-
-   if(eof && !in_buffer)
+   Init0();
+   InitVerify(f);
+}
+FileVerificator::FileVerificator(const FDStream *stream)
+{
+   Init0();
+   if(done)
+      return;
+   const char *f=stream->full_name;
+   if(!f)
    {
       done=true;
-      return MOVED;
+      return;
    }
-
-   int m=STALL;
-
-   if(!write_allowed)
-      return m;
-
-   while(in_buffer>0)
+   const char *cwd=stream->GetCwd();
+   int cwd_len=xstrlen(cwd);
+   if(cwd && cwd_len>0 && !strncmp(f,cwd,cwd_len))
    {
-      int res=Put_LL(buffer+buffer_ptr,in_buffer);
-      if(res>0)
+      f+=cwd_len;
+      while(*f=='/')
+	 f++;
+      if(*f==0)
+	 f=".";
+   }
+   InitVerify(f);
+   if(verify_process)
+   {
+      verify_process->SetProcGroup(stream->GetProcGroup());
+      verify_process->SetCwd(cwd);
+   }
+}
+FileVerificator::FileVerificator(FileAccess *session,const char *f)
+{
+   Init0();
+   if(done)
+      return;
+   if(strcmp(session->GetProto(),"file"))
+   {
+      done=true;
+      return;
+   }
+   InitVerify(f);
+   verify_process->SetCwd(session->GetCwd());
+}
+FileVerificator::~FileVerificator()
+{
+   xfree(error_text);
+   // verify_process is deleted by verify_buffer dtor
+   Delete(verify_buffer);
+}
+int FileVerificator::Do()
+{
+   int m=STALL;
+   if(done)
+      return m;
+   verify_process->Kill(SIGCONT);
+   if(!verify_buffer->Eof())
+      return m;
+   if(verify_process->GetProcState()!=ProcWait::TERMINATED)
+      return m;
+   done=true;
+   m=MOVED;
+   if(verify_process->GetProcExitCode()!=0)
+   {
+      verify_buffer->Put("",1);
+      int len;
+      const char *s;
+      verify_buffer->Get(&s,&len);
+      len=strlen(s);
+      while(len>0 && s[len-1]=='\n')
+	 len--;
+      if(len==0)
       {
-	 in_buffer-=res;
-	 buffer_ptr+=res;
-	 m=MOVED;
+	 s=_("Verify command failed without a message");
+	 len=strlen(s);
       }
-      if(res<0)
-	 return MOVED;
-      if(res==0)
-	 break;
+      error_text=xstrdup(s);
+      error_text[len]=0;
+      char *nl=strrchr(error_text,'\n');
+      if(nl)
+	 memmove(error_text,nl+1,len-(nl+1-error_text)+1);
    }
    return m;
 }
 
-void FileCopyPeerOutputJob::Fg()
-{
-   o->Fg();
-   FileCopyPeer::Fg();
-}
-void FileCopyPeerOutputJob::Bg()
-{
-   o->Bg();
-   FileCopyPeer::Bg();
-}
 
 // special pointer to creator of ftp/ftp copier. It is init'ed in Ftp class.
 FileCopy *(*FileCopy::fxp_create)(FileCopyPeer *src,FileCopyPeer *dst,bool cont);
