@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileSet.cc,v 1.50 2006/02/06 10:57:24 lav Exp $ */
+/* $Id: FileSet.cc,v 1.52.2.1 2007/04/05 13:23:37 lav Exp $ */
 
 #include <config.h>
 #include "FileSet.h"
@@ -42,11 +42,6 @@
 #include "IdNameCache.h"
 #include "PatternSet.h"
 
-#define NO_SIZE	     (-1L)
-#define NO_SIZE_YET  (-2L)
-#define NO_DATE	     ((time_t)-1L)
-#define NO_DATE_YET  ((time_t)-2L)
-
 void  FileInfo::Merge(const FileInfo& f)
 {
    if(strcmp(name,f.name))
@@ -55,8 +50,8 @@ void  FileInfo::Merge(const FileInfo& f)
    int dif=(~defined)&f.defined;
    if(dif&MODE)
       SetMode(f.mode);
-   if(dif&DATE || (defined&DATE && f.defined&DATE && f.date_prec<date_prec))
-      SetDate(f.date,f.date_prec);
+   if(dif&DATE || (defined&DATE && f.defined&DATE && f.date.ts_prec<date.ts_prec))
+      SetDate(f.date,f.date.ts_prec);
    if(dif&TYPE)
       SetType(f.filetype);
    if(dif&SYMLINK)
@@ -96,6 +91,17 @@ void FileInfo::SetGroup(const char *g)
    defined|=GROUP;
 }
 
+void FileSet::add_before(int pos,FileInfo *fi)
+{
+   if(fnum==allocated)
+   {
+      allocated=allocated?allocated*2:16;
+      files=files_sort=(FileInfo**)xrealloc(files,allocated*sizeof(*files));
+   }
+   memmove(files+pos+1, files+pos, sizeof(*files)*(fnum-pos));
+   files[pos]=fi;
+   fnum++;
+}
 void FileSet::Add(FileInfo *fi)
 {
    assert(!sorted);
@@ -111,14 +117,7 @@ void FileSet::Add(FileInfo *fi)
       delete fi;
       return;
    }
-   if(fnum==allocated)
-   {
-      allocated=allocated?allocated*2:16;
-      files=files_sort=(FileInfo**)xrealloc(files,allocated*sizeof(*files));
-   }
-   memmove(files+pos+1, files+pos, sizeof(*files)*(fnum-pos));
-   files[pos]=fi;
-   fnum++;
+   add_before(pos,fi);
 }
 
 void FileSet::Sub(int i)
@@ -134,21 +133,14 @@ void FileSet::Sub(int i)
 
 void FileSet::Merge(const FileSet *set)
 {
-   int i,j;
-   for(i=0; i<set->fnum; i++)
+   for(int i=0; i<set->fnum; i++)
    {
-      for(j=0; j<fnum; j++)
-      {
-      	 if(!strcmp(files[j]->name,set->files[i]->name))
-	 {
-	    files[j]->Merge(*(set->files[i]));
-	    break;
-	 }
-      }
-      if(j==fnum)
-      {
-	 Add(new FileInfo(*set->files[i]));
-      }
+      FileInfo *fi=set->files[i];
+      int pos = FindGEIndByName(fi->name);
+      if(pos < fnum && !strcmp(files[pos]->name,fi->name))
+	 files[pos]->Merge(*fi);
+      else
+	 add_before(pos,new FileInfo(*fi));
    }
 }
 
@@ -391,6 +383,21 @@ void FileSet::ExcludeDots()
    }
 }
 
+void FileSet::ExcludeUnaccessible()
+{
+   for(int i=0; i<fnum; i++)
+   {
+      if(!(files[i]->defined&FileInfo::MODE) || !(files[i]->defined&FileInfo::TYPE))
+	 continue;
+      if((files[i]->filetype==FileInfo::NORMAL    && !(files[i]->mode&0444))
+      || (files[i]->filetype==FileInfo::DIRECTORY && !(files[i]->mode&0444&(files[i]->mode<<2))))
+      {
+	 Sub(i);
+	 i--;
+      }
+   }
+}
+
 bool  FileInfo::SameAs(const FileInfo *fi,int ignore)
 {
    if(defined&NAME && fi->defined&NAME)
@@ -409,9 +416,9 @@ bool  FileInfo::SameAs(const FileInfo *fi,int ignore)
 
    if(defined&DATE && fi->defined&DATE && !(ignore&DATE))
    {
-      time_t p=date_prec;
-      if(p<fi->date_prec)
-	 p=fi->date_prec;
+      time_t p=date.ts_prec;
+      if(p<fi->date.ts_prec)
+	 p=fi->date.ts_prec;
       if(!(ignore&IGNORE_DATE_IF_OLDER && date<fi->date)
       && labs(date-fi->date)>p)
 	 return false;
@@ -576,7 +583,7 @@ void FileSet::LocalUtime(const char *dir,bool only_dirs)
 	 struct stat st;
 	 ut.actime=ut.modtime=file->date;
 
-	 if(stat(local_name,&st)!=-1 && st.st_mtime!=file->date)
+	 if(stat(local_name,&st)!=-1 && labs(st.st_mtime-file->date)>file->date.ts_prec)
 	    utime(local_name,&ut);
       }
    }
@@ -677,7 +684,6 @@ void FileInfo::Init()
    filetype=UNKNOWN;
    mode=(mode_t)-1;
    date=NO_DATE;
-   date_prec=0;
    size=NO_SIZE;
    nlinks=0;
    name=0;
@@ -700,7 +706,6 @@ FileInfo::FileInfo(const FileInfo &fi)
    filetype=fi.filetype;
    mode=fi.mode;
    date=fi.date;
-   date_prec=fi.date_prec;
    size=fi.size;
    nlinks=fi.nlinks;
    longname=xstrdup(fi.longname);
