@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: MirrorJob.cc,v 1.139 2006/08/29 07:25:26 lav Exp $ */
+/* $Id: MirrorJob.cc,v 1.139.2.2 2007/04/05 13:23:39 lav Exp $ */
 
 #include <config.h>
 
@@ -228,7 +228,7 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	    && (old->defined&file->TYPE) && old->filetype==old->NORMAL
 	    && (flags&IGNORE_TIME ||
 	    	((file->defined&file->DATE) && (old->defined&old->DATE)
-	    	&& file->date + file->date_prec < old->date - old->date_prec))
+	    	&& file->date + file->date.ts_prec < old->date - old->date.ts_prec))
 	    && (file->defined&file->SIZE) && (old->defined&old->SIZE)
 	    && file->size >= old->size
 	    && !use_pget)
@@ -297,7 +297,7 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	       c->RemoveTargetFirst();
 	    CopyJob *cp=
 	       new CopyJob(c,file->name,"mirror");
-	    if((file->defined&file->DATE) && file->date_prec<=1)
+	    if(file->defined&file->DATE)
 	       cp->SetDate(file->date);
 	    if(file->defined&file->SIZE)
 	       cp->SetSize(file->size);
@@ -419,8 +419,8 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	    ArgV args("shell");
 	    args.Append("ln");
 	    args.Append("-sf");
-	    args.Append(file->symlink);
-	    args.Append(target_name);
+	    args.Append(shell_encode(file->symlink));
+	    args.Append(shell_encode(target_name));
 	    char *cmd=args.CombineQuoted();
 	    fprintf(script,"%s\n",cmd);
 	    xfree(cmd);
