@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: NetAccess.cc,v 1.86 2006/06/15 14:18:40 lav Exp $ */
+/* $Id: NetAccess.cc,v 1.87.2.1 2007/04/05 13:23:39 lav Exp $ */
 
 #include <config.h>
 
@@ -814,6 +814,7 @@ do_again:
       FileSet *set=Parse(b,len);
       if(set)
       {
+	 bool need_resort=false;
 	 set->rewind();
 	 for(file=set->curr(); file!=0; file=set->next())
 	 {
@@ -824,8 +825,11 @@ do_again:
 	       file->name=(char*)xrealloc(file->name,len+3);
 	       memmove(file->name+2,file->name,len+1);
 	       memcpy(file->name,"./",2);
+	       need_resort=true;
 	    }
 	 }
+	 if(need_resort && !result)
+	    result=new FileSet; // Merge will sort the names
 	 if(result)
 	 {
 	    result->Merge(set);
@@ -867,7 +871,7 @@ do_again:
 	 cur->time=NO_DATE;
 	 cur->get_size = need_size && !(file->defined & file->SIZE);
 	 cur->get_time = need_time && (!(file->defined & file->DATE)
-				 || (file->date_prec>0 && can_get_prec_time));
+				 || (file->date.ts_prec>0 && can_get_prec_time));
 	 cur->file=0;
 
 	 if(file->defined & file->TYPE)
