@@ -24,7 +24,7 @@
 #include "Job.h"
 #include "StatusLine.h"
 #include "ArgV.h"
-#include <stdio.h>
+#include "trio.h"
 
 class mvJob : public SessionJob
 {
@@ -34,7 +34,7 @@ public:
    int	 Done() { return session->IsClosed(); }
    int	 ExitCode() { return failed; }
 
-   void	 PrintStatus(int);
+   void	 PrintStatus(int,const char *);
    void	 ShowRunStatus(StatusLine *);
    void	 SayFinal();
 
